print("import: 'surprise'")
import surprise

print("import: 'surprise.prediction_algorithms'")
import surprise.prediction_algorithms

print("import: 'surprise.similarities'")
import surprise.similarities

print("import: 'surprise.dataset'")
import surprise.dataset

print("import: 'surprise.model_selection'")
import surprise.model_selection

